                     /*-----------------------------------*/
                     /*             >>>Slip<<<            */
                     /*            Theo D'Hondt           */
                     /*     VUB Software Languages Lab    */
                     /*             (c) 2010              */
                     /*-----------------------------------*/
                     /*   version 9: garbage collection   */
                     /*-----------------------------------*/
                     /*              natives              */
                     /*-----------------------------------*/

#include <errno.h>
#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>

#include "SlipGrammar.h"

#include "SlipCompile.h"
#include "SlipDictionary.h"
#include "SlipEnvironment.h"
#include "SlipEvaluate.h"
#include "SlipPool.h"
#include "SlipPrint.h"
#include "SlipRead.h"
#include "SlipStack.h"
#include "SlipThread.h"

#include "SlipNative.h"

/*------------------------------- private macros -----------------------------*/

#define NULL_PAIR (PAI_type)Grammar_Null

/*----------------------------- private constants ----------------------------*/

static const FLO_type Epsilon = 1.0E-8;

static const TXT_type AL1_message = "at least one argument required";
static const TXT_type AL2_message = "at least two arguments required";
static const TXT_type AM1_message = "at most one argument required";
static const TXT_type AM2_message = "at most two arguments required";
static const TXT_type ARN_message = "arguments restricted to numbers";
static const TXT_type DBZ_message = "divide by zero";
static const TXT_type EX0_message = "no arguments allowed";
static const TXT_type EX1_message = "exactly one argument required";
static const TXT_type EX2_message = "exactly two arguments required";
static const TXT_type EX3_message = "exactly three arguments required";
static const TXT_type IOR_message = "index out of range";
static const TXT_type NAL_message = "not a list";
static const TXT_type NAN_message = "not a number";
static const TXT_type NAP_message = "not a pair";
static const TXT_type NAS_message = "not a string";
static const TXT_type NAV_message = "not a vector";
static const TXT_type NOR_message = "number out of range";
static const TXT_type NNN_message = "non-negative number required";

/*----------------------------- private functions ----------------------------*/

static VOI_type native_define_M(TXT_type Name,
                                FUN_type Function)
  { EXP_type native;
    SYM_type variable;
    UNS_type raw_offset;
    variable = Pool_Enter_C(Name);
    native = make_NAT_M(Function,
                        Name);
    raw_offset = Dictionary_Define_Symbol_M(variable,
                                            FALSE);
    Pool_Enter_Native_Name(Name,
                           raw_offset);
    Environment_Local_Set(raw_offset,
                          native); }

static EXP_type error(TXT_type Message)
  { Main_Error(Message);
    return Grammar_Unspecified; }

static EXP_type error_text(TXT_type Message,
                           TXT_type Text )
  { sprintf(Main_Buffer, "%s: %s", Message, Text);
    return error(Main_Buffer); }

/*----------------------------------------------------------------------------*/
/*-------------------------------- arithmetic --------------------------------*/
/*----------------------------------------------------------------------------*/

/*------------------------------------- + ------------------------------------*/

static const TXT_type add_text = "+";

static EXP_type add_floats_C(FLO_type Raw_real,
                             VEC_type Frame,
                             UNS_type Index)
  { EXP_type value;
    FLO_type raw_real,
             raw_real_sum;
    REA_type real;
    SGN_type raw_number;
    TAG_type tag;
    UNS_type index,
             size;
    raw_real_sum = Raw_real;
    size = size_VEC(Frame);
    for (index = Index;
         index <= size;
         index++)
      { value = Frame[index];
        tag = Grammar_Tag(value);
        switch (tag)
          { case NBR_tag:
              raw_number = get_NBR(value);
              raw_real_sum += raw_number;
              break;
            case REA_tag:
              real = value;
              raw_real = real->flo;
              raw_real_sum += raw_real;
              break;
            default:
              return error_text(ARN_message,
                                add_text); }}
    Grammar_Default_Claim_C();
    return make_REA_M(raw_real_sum); }

static EXP_type add_integers_C(NBR_type Number,
                               VEC_type Frame)
  { DBL_type raw_number_sum;
    EXP_type value;
    FLO_type raw_real,
             raw_real_sum;
    REA_type real;
    SGN_type raw_number;
    TAG_type tag;
    UNS_type index,
             size;
    raw_number_sum = get_NBR(Number);
    size = size_VEC(Frame);
    for (index = 2;
         index <= size;
         index++)
      { value = Frame[index];
        tag = Grammar_Tag(value);
        switch (tag)
          { case NBR_tag:
              raw_number = get_NBR(value);
              raw_number_sum += raw_number;
              break;
            case REA_tag:
              real = value;
              raw_real = real->flo;
              raw_real_sum = raw_number_sum + raw_real;
              return add_floats_C(raw_real_sum,
                                  Frame,
                                  index + 1);
            default:
              return error_text(ARN_message,
                                add_text); }}
    return Grammar_Number(raw_number_sum); }

static EXP_type add_native_C(VEC_type Frame)
  { EXP_type value;
    FLO_type raw_real;
    REA_type real;
    TAG_type tag;
    UNS_type size;
    size = size_VEC(Frame);
    if (size == 0)
      return Grammar_Zero;
    value = Frame[1];
    tag = Grammar_Tag(value);
    switch (tag)
      { case NBR_tag:
          if (size == 1)
            return value;
          return add_integers_C(value,
                                Frame);
        case REA_tag:
          if (size == 1)
            return value;
          real = value;
          raw_real = real->flo;
          return add_floats_C(raw_real,
                              Frame,
                              2);
        default:
          return error_text(ARN_message,
                            add_text); }}

static VOI_type initialize_add_native_M(VOI_type)
  { native_define_M(add_text,
                    add_native_C); }

/*------------------------------------- - ------------------------------------*/

static const TXT_type sub_text = "-";

static EXP_type subtract_floats_C(FLO_type Raw_real,
                                  VEC_type Frame,
                                  UNS_type Index)
  { EXP_type value;
    FLO_type raw_real,
             raw_real_sum;
    REA_type real;
    SGN_type raw_number;
    TAG_type tag;
    UNS_type index,
             size;
    raw_real_sum = Raw_real;
    size = size_VEC(Frame);
    for (index = Index;
         index <= size;
         index++)
      { value = Frame[index];
        tag = Grammar_Tag(value);
        switch (tag)
          { case NBR_tag:
              raw_number = get_NBR(value);
              raw_real_sum -= raw_number;
              break;
            case REA_tag:
              real = value;
              raw_real = real->flo;
              raw_real_sum -= raw_real;
              break;
            default:
              return error_text(ARN_message,
                                sub_text); }}
    Grammar_Default_Claim_C();
    return make_REA_M(raw_real_sum); }

static EXP_type subtract_integers_C(NBR_type Number,
                                    VEC_type Frame)
  { DBL_type raw_number_sum;
    EXP_type value;
    FLO_type raw_real,
             raw_real_sum;
    REA_type real;
    SGN_type raw_number;
    TAG_type tag;
    UNS_type index,
             size;
    raw_number_sum = get_NBR(Number);
    size = size_VEC(Frame);
    for (index = 2;
         index <= size;
         index++)
      { value = Frame[index];
        tag = Grammar_Tag(value);
        switch (tag)
          { case NBR_tag:
              raw_number = get_NBR(value);
              raw_number_sum -= raw_number;
              break;
            case REA_tag:
              real = value;
              raw_real = real->flo;
              raw_real_sum = raw_number_sum - raw_real;
              return subtract_floats_C(raw_real_sum,
                                       Frame,
                                       index + 1);
            default:
              return error_text(ARN_message,
                                sub_text); }}
    return Grammar_Number(raw_number_sum); }

static EXP_type subtract_native_C(VEC_type Frame)
  { EXP_type value;
    FLO_type raw_real;
    REA_type real;
    SGN_type raw_number;
    TAG_type tag;
    UNS_type size;
    size = size_VEC(Frame);
    if (size == 0)
      return error_text(AL1_message,
                        sub_text);
    value = Frame[1];
    tag = Grammar_Tag(value);
    if (size == 1)
      { switch (tag)
          { case NBR_tag:
              raw_number  = get_NBR(value);
              return make_NBR(- raw_number);
            case REA_tag:
              real = value;
              raw_real = real->flo;
              Grammar_Default_Claim_C();
              return make_REA_M(- raw_real);
            default:
              return error_text(ARN_message,
                                sub_text); }}
    switch (tag)
      { case NBR_tag:
          return subtract_integers_C(value,
                                     Frame);
        case REA_tag:
          real = value;
          raw_real = real->flo;
          return subtract_floats_C(raw_real,
                                   Frame,
                                   2);
        default:
          return error_text(ARN_message,
                            sub_text); }}

static VOI_type initialize_subtract_native_M(VOI_type)
  { native_define_M(sub_text,
                    subtract_native_C); }

/*------------------------------------- * ------------------------------------*/

static const TXT_type mul_text = "*";

static EXP_type multiply_floats_C(FLO_type Raw_real,
                                  VEC_type Frame,
                                  UNS_type Index)
  { EXP_type value;
    FLO_type raw_real,
             raw_real_product;
    REA_type real;
    SGN_type raw_number;
    TAG_type tag;
    UNS_type index,
             size;
    raw_real_product = Raw_real;
    size = size_VEC(Frame);
    for (index = Index;
         index <= size;
         index++)
      { value = Frame[index];
        tag   = Grammar_Tag(value);
        switch (tag)
          { case NBR_tag:
              raw_number = get_NBR(value);
              raw_real_product *= raw_number;
              break;
            case REA_tag:
              real = value;
              raw_real = real->flo;
              raw_real_product *= raw_real;
              break;
            default:
              return error_text(ARN_message,
                                mul_text); }}
    Grammar_Default_Claim_C();
    return make_REA_M(raw_real_product); }

static EXP_type multiply_integers_C(NBR_type Number,
                                    VEC_type Frame)
  { DBL_type raw_number_product;
    EXP_type value;
    FLO_type raw_real,
             raw_real_product;
    REA_type real;
    SGN_type raw_number;
    TAG_type tag;
    UNS_type index,
             size;
    raw_number_product = get_NBR(Number);
    size = size_VEC(Frame);
    for (index = 2;
         index <= size;
         index++)
      { value  = Frame[index];
        tag = Grammar_Tag(value);
        switch (tag)
          { case NBR_tag:
              raw_number = get_NBR(value);
              raw_number_product *= raw_number;
              if (errno == ERANGE)
                return error_text(NOR_message,
                                  mul_text);
              break;
            case REA_tag:
              real = value;
              raw_real = real->flo;
              raw_real_product = raw_number_product * raw_real;
              return multiply_floats_C(raw_real_product,
                                       Frame,
                                       index + 1);
            default:
              return error_text(ARN_message,
                                mul_text); }}
    return Grammar_Number(raw_number_product); }

static EXP_type multiply_native_C(VEC_type Frame)
  { EXP_type value;
    FLO_type raw_real;
    REA_type real;
    TAG_type tag;
    UNS_type size;
    size = size_VEC(Frame);
    if (size == 0)
      return Grammar_One;
    value = Frame[1];
    if (size == 1)
      return value;
    tag = Grammar_Tag(value);
    switch (tag)
      { case NBR_tag:
          return multiply_integers_C(value,
                                     Frame);
        case REA_tag:
          real = value;
          raw_real = real->flo;
          return multiply_floats_C(raw_real,
                                   Frame,
                                   2);
        default:
          return error_text(ARN_message,
                            mul_text); }}

static VOI_type initialize_multiply_native_M(VOI_type)
  { native_define_M(mul_text,
                    multiply_native_C); }

/*------------------------------------- / ------------------------------------*/

static const TXT_type div_text = "/";

static EXP_type divide_floats_C(FLO_type Raw_real,
                                VEC_type Frame)
  { EXP_type value;
    FLO_type raw_real,
             raw_real_quotient;
    REA_type real;
    SGN_type raw_number;
    TAG_type tag;
    UNS_type index,
             size;
    raw_real_quotient = Raw_real;
    size = size_VEC(Frame);
    for (index = 2;
         index <= size;
         index++)
      { value = Frame[index];
        tag = Grammar_Tag(value);
        switch (tag)
          { case NBR_tag:
              raw_number = get_NBR(value);
              if (raw_number == 0)
                return error_text(DBZ_message,
                                  div_text);
              raw_real_quotient /= raw_number;
              break;
            case REA_tag:
              real = value;
              raw_real = real->flo;
              if (fabs(raw_real) < Epsilon)
                return error_text(DBZ_message,
                                  div_text);
              raw_real_quotient /= raw_real;
              break;
            default:
              return error_text(ARN_message,
                                div_text); }}
    Grammar_Default_Claim_C();
    return make_REA_M(raw_real_quotient); }

static EXP_type divide_native_C(VEC_type Frame)
  { EXP_type value;
    FLO_type raw_real;
    REA_type real;
    SGN_type raw_number;
    TAG_type tag;
    UNS_type size;
    size = size_VEC(Frame);
    if (size == 0)
      return error_text(AL1_message,
                        div_text);
    value = Frame[1];
    tag = Grammar_Tag(value);
    if (size == 1)
      { switch (tag)
          { case NBR_tag:
              raw_number = get_NBR(value);
              if (raw_number == 0)
                return error_text(DBZ_message,
                                  div_text);
              Grammar_Default_Claim_C();
              return make_REA_M(1.0 / raw_number);
            case REA_tag:
              real = value;
              raw_real = real->flo;
              if (fabs(raw_real) < Epsilon)
                return error_text(DBZ_message,
                                  div_text);
              Grammar_Default_Claim_C();
              return make_REA_M(1.0 / raw_real);
            default:
              return error_text(ARN_message,
                                div_text); }}
    switch (tag)
      { case NBR_tag:
          raw_number = get_NBR(value);
          return divide_floats_C(raw_number,
                                 Frame);
        case REA_tag:
          real = value;
          raw_real = real->flo;
          return divide_floats_C(raw_real,
                                 Frame);
        default:
          return error_text(ARN_message,
                            div_text); }}

static VOI_type initialize_divide_native_M(VOI_type)
  { native_define_M(div_text,
                    divide_native_C); }

/*--------------------------------- quotient ---------------------------------*/

static const TXT_type quo_text = "quotient";

static EXP_type quotient_native(VEC_type Frame)
  { DBL_type raw_number;
    NBR_type left,
             number,
             right;
    SGN_type raw_left,
             raw_right;
    UNS_type size;
    size = size_VEC(Frame);
    if (size != 2)
      return error_text(EX2_message,
                        quo_text);
    left  = Frame[1];
    right = Frame[2];
    if (!is_NBR(left) || !is_NBR(right))
      return error_text(ARN_message,
                        quo_text);
    raw_left = get_NBR(left);
    raw_right = get_NBR(right);
    if (raw_right == 0)
      return error_text(DBZ_message,
                        quo_text);
    raw_number = raw_left / raw_right;
    number = Grammar_Number(raw_number);
    return number; }

static VOI_type initialize_quotient_native_M(VOI_type)
  { native_define_M(quo_text,
                    quotient_native); }

/*-------------------------------- remainder ---------------------------------*/

static const TXT_type rem_text = "remainder";

static EXP_type remainder_native(VEC_type Frame)
  { DBL_type raw_number;
    NBR_type left,
             number,
             right;
    SGN_type raw_left,
             raw_right;
    UNS_type size;
    size = size_VEC(Frame);
    if (size != 2)
      return error_text(EX2_message,
                        rem_text);
    left  = Frame[1];
    right = Frame[2];
    if (!is_NBR(left) || !is_NBR(right))
      return error_text(ARN_message,
                        rem_text);
    raw_left = get_NBR(left);
    raw_right = get_NBR(right);
    if (raw_right == 0)
      return error_text(DBZ_message,
                        rem_text);
    raw_number = raw_left % raw_right;
    number = Grammar_Number(raw_number);
    return number; }

static VOI_type initialize_remainder_native_M(VOI_type)
  { native_define_M(rem_text,
                    remainder_native); }

/*----------------------------------------------------------------------------*/
/*------------------------- transcendental functions -------------------------*/
/*----------------------------------------------------------------------------*/

/*----------------------------------- sqrt -----------------------------------*/

static const TXT_type sqt_text = "sqrt";

static EXP_type sqrt_native_C(VEC_type Frame)
  { EXP_type value;
    FLO_type raw_float;
    REA_type real;
    SGN_type raw_number;
    TAG_type tag;
    UNS_type size;
    size = size_VEC(Frame);
    if (size != 1)
      return error_text(EX1_message,
                        sqt_text);
    value = Frame[1];
    tag = Grammar_Tag(value);
    switch (tag)
      { case NBR_tag:
          raw_number = get_NBR(value);
          raw_float = sqrt(raw_number);
          break;
        case REA_tag:
          real = value;
          raw_float = real->flo;
          raw_float = sqrt(raw_float);
          break;
        default:
          return error_text(ARN_message,
                            sqt_text); }
    Grammar_Default_Claim_C();
    return make_REA_M(raw_float); }

static VOI_type initialize_sqrt_native_M(VOI_type)
  { native_define_M(sqt_text,
                    sqrt_native_C); }

/*----------------------------------------------------------------------------*/
/*-------------------------------- comparison --------------------------------*/
/*----------------------------------------------------------------------------*/

/*------------------------------------- = ------------------------------------*/

static const TXT_type eql_text = "=";

static EXP_type equal_floats(FLO_type Raw_real,
                             VEC_type Frame,
                             UNS_type Index)
  { EXP_type value;
    FLO_type raw_previous_real,
             raw_real;
    REA_type real;
    TAG_type tag;
    UNS_type index,
             size;
    raw_previous_real = Raw_real;
    size = size_VEC(Frame);
    for (index = Index;
         index <= size;
         index++)
      { value = Frame[index];
        tag = Grammar_Tag(value);
        switch (tag)
          { case NBR_tag:
              raw_real = get_NBR(value);
              break;
            case REA_tag:
              real = value;
              raw_real = real->flo;
              break;
            default:
              return error_text(ARN_message,
                                eql_text); }
        if (raw_previous_real != raw_real)
          return Grammar_False;
        raw_previous_real = raw_real; }
   return Grammar_True; }

static EXP_type equal_integers(NBR_type Number,
                               VEC_type Frame)
  { EXP_type value;
    FLO_type raw_real;
    REA_type real;
    SGN_type raw_number,
             raw_previous_number;
    TAG_type tag;
    UNS_type index,
             size;
    raw_previous_number = get_NBR(Number);
    size = size_VEC(Frame);
    for (index = 2;
         index <= size;
         index++)
      { value = Frame[index];
        tag = Grammar_Tag(value);
        switch (tag)
          { case NBR_tag:
              raw_number = get_NBR(value);
              if (raw_previous_number != raw_number)
                return Grammar_False;
              break;
            case REA_tag:
              real = value;
              raw_real = real->flo;
              if (raw_previous_number != raw_real)
                return Grammar_False;
              return equal_floats(raw_real,
                                  Frame,
                                  index + 1);
            default:
              return error_text(ARN_message,
                                eql_text); }
        raw_previous_number = raw_number; }
    return Grammar_True; }

static EXP_type equal_native(VEC_type Frame)
  { EXP_type value;
    FLO_type raw_real;
    REA_type real;
    TAG_type tag;
    UNS_type size;
    size = size_VEC(Frame);
    if (size < 2)
      return error_text(AL2_message,
                        eql_text);
    value = Frame[1];
    tag = Grammar_Tag(value);
    switch (tag)
      { case NBR_tag:
          return equal_integers(value,
                                Frame);
        case REA_tag:
          real = value;
          raw_real = real->flo;
          return equal_floats(raw_real,
                              Frame,
                              2);
        default:
          return error_text(ARN_message,
                            eql_text); }}

static VOI_type initialize_equal_native_M(VOI_type)
  { native_define_M(eql_text,
                    equal_native); }

/*------------------------------------- > ------------------------------------*/

static const TXT_type grt_text = ">";

static EXP_type greater_floats(FLO_type Raw_real,
                               VEC_type Frame,
                               UNS_type Index)
  { EXP_type value;
    FLO_type raw_previous_real,
             raw_real;
    REA_type real;
    TAG_type tag;
    UNS_type index,
             size;
    raw_previous_real = Raw_real;
    size = size_VEC(Frame);
    for (index = Index;
         index <= size;
         index++)
      { value = Frame[index];
        tag = Grammar_Tag(value);
        switch (tag)
          { case NBR_tag:
              raw_real = get_NBR(value);
              break;
            case REA_tag:
              real = value;
              raw_real = real->flo;
              break;
            default:
              return error_text(ARN_message,
                                grt_text); }
        if (raw_previous_real <= raw_real)
          return Grammar_False;
        raw_previous_real = raw_real; }
   return Grammar_True; }

static EXP_type greater_integers(NBR_type Number,
                                 VEC_type Frame)
  { EXP_type value;
    FLO_type raw_real;
    REA_type real;
    SGN_type raw_number,
             raw_previous_number;
    TAG_type tag;
    UNS_type index,
             size;
    raw_previous_number = get_NBR(Number);
    size = size_VEC(Frame);
    for (index = 2;
         index <= size;
         index++)
      { value = Frame[index];
        tag = Grammar_Tag(value);
        switch (tag)
          { case NBR_tag:
              raw_number = get_NBR(value);
              if (raw_previous_number <= raw_number)
                return Grammar_False;
              break;
            case REA_tag:
              real = value;
              raw_real = real->flo;
              if (raw_previous_number <= raw_real)
                return Grammar_False;
              return greater_floats(raw_real,
                                    Frame,
                                    index + 1);
            default:
              return error_text(ARN_message,
                                grt_text); }
        raw_previous_number = raw_number; }
    return Grammar_True; }

static EXP_type greater_native(VEC_type Frame)
  { EXP_type value;
    FLO_type raw_real;
    REA_type real;
    TAG_type tag;
    UNS_type size;
    size = size_VEC(Frame);
    if (size < 2)
      return error_text(AL2_message,
                        grt_text);
    value = Frame[1];
    tag = Grammar_Tag(value);
    switch (tag)
      { case NBR_tag:
          return greater_integers(value,
                                  Frame);
        case REA_tag:
          real = value;
          raw_real = real->flo;
          return greater_floats(raw_real,
                                Frame,
                                2);
        default:
          return error_text(ARN_message,
                            grt_text); }}

static VOI_type initialize_greater_native_M(VOI_type)
  { native_define_M(grt_text,
                    greater_native); }

/*------------------------------------- < ------------------------------------*/

static const TXT_type lss_text = "<";

static EXP_type less_floats(FLO_type Raw_real,
                            VEC_type Frame,
                            UNS_type Index)
  { EXP_type value;
    FLO_type raw_previous_real,
             raw_real;
    REA_type real;
    TAG_type tag;
    UNS_type index,
             size;
    raw_previous_real = Raw_real;
    size = size_VEC(Frame);
    for (index = Index;
         index <= size;
         index++)
      { value = Frame[index];
        tag = Grammar_Tag(value);
        switch (tag)
          { case NBR_tag:
              raw_real = get_NBR(value);
              break;
            case REA_tag:
              real = value;
              raw_real = real->flo;
              break;
            default:
              return error_text(ARN_message,
                                lss_text); }
        if (raw_previous_real >= raw_real)
          return Grammar_False;
        raw_previous_real = raw_real; }
   return Grammar_True; }

static EXP_type less_integers(NBR_type Number,
                              VEC_type Frame)
  { EXP_type value;
    FLO_type raw_real;
    REA_type real;
    SGN_type raw_number,
             raw_previous_number;
    TAG_type tag;
    UNS_type index,
             size;
    raw_previous_number = get_NBR(Number);
    size = size_VEC(Frame);
    for (index = 2;
         index <= size;
         index++)
      { value = Frame[index];
        tag = Grammar_Tag(value);
        switch (tag)
          { case NBR_tag:
              raw_number = get_NBR(value);
              if (raw_previous_number >= raw_number)
                return Grammar_False;
              break;
            case REA_tag:
              real = value;
              raw_real = real->flo;
              if (raw_previous_number >= raw_real)
                return Grammar_False;
              return greater_floats(raw_real,
                                    Frame,
                                    index + 1);
            default:
              return error_text(ARN_message,
                                lss_text); }
        raw_previous_number = raw_number; }
    return Grammar_True; }

static EXP_type less_native(VEC_type Frame)
  { EXP_type value;
    FLO_type raw_real;
    REA_type real;
    TAG_type tag;
    UNS_type size;
    size = size_VEC(Frame);
    if (size < 2)
      return error_text(AL2_message,
                        lss_text);
    value = Frame[1];
    tag = Grammar_Tag(value);
    switch (tag)
      { case NBR_tag:
          return less_integers(value,
                               Frame);
        case REA_tag:
          real = value;
          raw_real = real->flo;
          return less_floats(raw_real,
                             Frame,
                             2);
        default:
          return error_text(ARN_message,
                            lss_text); }}

static VOI_type initialize_less_native_M(VOI_type)
  { native_define_M(lss_text,
                    less_native); }

/*------------------------------------- >= -----------------------------------*/

static const TXT_type geq_text = ">=";

static EXP_type greater_or_equal_floats(FLO_type Raw_real,
                                        VEC_type Frame,
                                        UNS_type Index)
  { EXP_type value;
    FLO_type raw_previous_real,
             raw_real;
    REA_type real;
    TAG_type tag;
    UNS_type index,
             size;
    raw_previous_real = Raw_real;
    size = size_VEC(Frame);
    for (index = Index;
         index <= size;
         index++)
      { value = Frame[index];
        tag = Grammar_Tag(value);
        switch (tag)
          { case NBR_tag:
              raw_real = get_NBR(value);
              break;
            case REA_tag:
              real = value;
              raw_real = real->flo;
              break;
            default:
              return error_text(ARN_message,
                                geq_text); }
        if (raw_previous_real < raw_real)
          return Grammar_False;
        raw_previous_real = raw_real; }
   return Grammar_True; }

static EXP_type greater_or_equal_integers(NBR_type Number,
                                          VEC_type Frame)
  { EXP_type value;
    FLO_type raw_real;
    REA_type real;
    SGN_type raw_number,
             raw_previous_number;
    TAG_type tag;
    UNS_type index,
             size;
    raw_previous_number = get_NBR(Number);
    size = size_VEC(Frame);
    for (index = 2;
         index <= size;
         index++)
      { value = Frame[index];
        tag = Grammar_Tag(value);
        switch (tag)
          { case NBR_tag:
              raw_number = get_NBR(value);
              if (raw_previous_number < raw_number)
                return Grammar_False;
              break;
            case REA_tag:
              real = value;
              raw_real = real->flo;
              if (raw_previous_number < raw_real)
                return Grammar_False;
              return greater_floats(raw_real,
                                    Frame,
                                    index + 1);
            default:
              return error_text(ARN_message,
                                geq_text); }
        raw_previous_number = raw_number; }
    return Grammar_True; }

static EXP_type greater_or_equal_native(VEC_type Frame)
  { EXP_type value;
    FLO_type raw_real;
    REA_type real;
    TAG_type tag;
    UNS_type size;
    size = size_VEC(Frame);
    if (size < 2)
      return error_text(AL2_message,
                        geq_text);
    value = Frame[1];
    tag = Grammar_Tag(value);
    switch (tag)
      { case NBR_tag:
          return greater_or_equal_integers(value,
                                           Frame);
        case REA_tag:
          real = value;
          raw_real = real->flo;
          return greater_or_equal_floats(raw_real,
                                         Frame,
                                         2);
        default:
          return error_text(ARN_message,
                            geq_text); }}

static VOI_type initialize_greater_or_equal_native_M(VOI_type)
  { native_define_M(geq_text,
                    greater_or_equal_native); }

/*------------------------------------- <= -----------------------------------*/

static const TXT_type leq_text = "<=";

static EXP_type less_or_equal_floats(FLO_type Raw_real,
                                     VEC_type Frame,
                                     UNS_type Index)
  { EXP_type value;
    FLO_type raw_previous_real,
             raw_real;
    REA_type real;
    TAG_type tag;
    UNS_type index,
             size;
    raw_previous_real = Raw_real;
    size = size_VEC(Frame);
    for (index = Index;
         index <= size;
         index++)
      { value = Frame[index];
        tag = Grammar_Tag(value);
        switch (tag)
          { case NBR_tag:
              raw_real = get_NBR(value);
              break;
            case REA_tag:
              real = value;
              raw_real = real->flo;
              break;
            default:
              return error_text(ARN_message,
                                leq_text); }
        if (raw_previous_real > raw_real)
          return Grammar_False;
        raw_previous_real = raw_real; }
   return Grammar_True; }

static EXP_type less_or_equal_integers(NBR_type Number,
                                       VEC_type Frame)
  { EXP_type value;
    FLO_type raw_real;
    REA_type real;
    SGN_type raw_number,
             raw_previous_number;
    TAG_type tag;
    UNS_type index,
             size;
    raw_previous_number = get_NBR(Number);
    size = size_VEC(Frame);
    for (index = 2;
         index <= size;
         index++)
      { value = Frame[index];
        tag = Grammar_Tag(value);
        switch (tag)
          { case NBR_tag:
              raw_number = get_NBR(value);
              if (raw_previous_number > raw_number)
                return Grammar_False;
              break;
            case REA_tag:
              real = value;
              raw_real = real->flo;
              if (raw_previous_number > raw_real)
                return Grammar_False;
              return greater_floats(raw_real,
                                    Frame,
                                    index + 1);
            default:
              return error_text(ARN_message,
                                grt_text); }
        raw_previous_number = raw_number; }
    return Grammar_True; }

static EXP_type less_or_equal_native(VEC_type Frame)
  { EXP_type value;
    FLO_type raw_real;
    REA_type real;
    TAG_type tag;
    UNS_type size;
    size = size_VEC(Frame);
    if (size < 2)
      return error_text(AL2_message,
                        leq_text);
    value = Frame[1];
    tag = Grammar_Tag(value);
    switch (tag)
      { case NBR_tag:
          return less_or_equal_integers(value,
                                        Frame);
        case REA_tag:
          real = value;
          raw_real = real->flo;
          return less_or_equal_floats(raw_real,
                                      Frame,
                                      2);
        default:
          return error_text(ARN_message,
                            leq_text); }}

static VOI_type initialize_less_or_equal_native_M(VOI_type)
  { native_define_M(leq_text,
                    less_or_equal_native); }

/*------------------------------------ eq? -----------------------------------*/

static const TXT_type eqq_text = "eq?";

static EXP_type equivalent_native(VEC_type Frame)
  { ADR_type left,
             right;
    UNS_type size;
    size = size_VEC(Frame);
    if (size != 2)
      return error_text(EX2_message,
                        eqq_text);
    left  = Frame[1];
    right = Frame[2];
    if (left == right)
      return Grammar_True;
    return Grammar_False; }

static VOI_type initialize_equivalent_native_M(VOI_type)
  { native_define_M(eqq_text,
                    equivalent_native); }

/*----------------------------------------------------------------------------*/
/*--------------------------------- booleans ---------------------------------*/
/*----------------------------------------------------------------------------*/

/*------------------------------------ not -----------------------------------*/

static const TXT_type not_text = "not";

static EXP_type not_native(VEC_type Frame)
  { EXP_type value;
    UNS_type size;
    size = size_VEC(Frame);
    if (size != 1)
      return error_text(EX1_message,
                        not_text);
    value = Frame[1];
    if (is_FLS(value))
      return Grammar_True;
    return Grammar_False; }

static VOI_type initialize_not_native_M(VOI_type)
  { native_define_M(not_text,
                    not_native); }

/*----------------------------------------------------------------------------*/
/*------------------------------- predicates ---------------------------------*/
/*----------------------------------------------------------------------------*/

/*---------------------------------- null? -----------------------------------*/

static const TXT_type inl_text = "null?";

static EXP_type is_null_native(VEC_type Frame)
  { EXP_type expression;
    UNS_type size;
    size = size_VEC(Frame);
    if (size != 1)
      return error_text(EX1_message,
                        inl_text);
    expression = Frame[1];
    if (is_NUL(expression))
      return Grammar_True;
    return Grammar_False; }

static VOI_type initialize_is_null_native_M(VOI_type)
  { native_define_M(inl_text,
                    is_null_native); }

/*---------------------------------- pair? -----------------------------------*/

static const TXT_type ipr_text = "pair?";

static EXP_type is_pair_native(VEC_type Frame)
  { EXP_type expression;
    UNS_type size;
    size = size_VEC(Frame);
    if (size != 1)
      return error_text(EX1_message,
                        ipr_text);
    expression = Frame[1];
    if (is_PAI(expression))
      return Grammar_True;
    return Grammar_False; }

static VOI_type initialize_is_pair_native_M(VOI_type)
  { native_define_M(ipr_text,
                    is_pair_native); }

/*------------------------------- procedure? ---------------------------------*/

static const TXT_type ipc_text = "procedure?";

static EXP_type is_procedure_native(VEC_type Frame)
  { EXP_type expression;
    UNS_type size;
    size = size_VEC(Frame);
    if (size != 1)
      return error_text(EX1_message,
                        ipr_text);
    expression = Frame[1];
    if (is_PRC(expression))
      return Grammar_True;
    if (is_PRZ(expression))
      return Grammar_True;
    if (is_NAT(expression))
      return Grammar_True;
    return Grammar_False; }

static VOI_type initialize_is_procedure_native_M(VOI_type)
  { native_define_M(ipc_text,
                    is_procedure_native); }

/*----------------------------------- real? ----------------------------------*/

static const TXT_type isr_text = "real?";

static EXP_type is_real_native(VEC_type Frame)
  { EXP_type expression;
    UNS_type size;
    size = size_VEC(Frame);
    if (size != 1)
      return error_text(EX1_message,
                        isr_text);
    expression = Frame[1];
    if (is_SYM(expression))
      return Grammar_True;
    return Grammar_False; }

static VOI_type initialize_is_real_native_M(VOI_type)
  { native_define_M(isr_text,
                    is_real_native); }

/*--------------------------------- symbol? ----------------------------------*/

static const TXT_type ism_text = "symbol?";

static EXP_type is_symbol_native(VEC_type Frame)
  { EXP_type expression;
    UNS_type size;
    size = size_VEC(Frame);
    if (size != 1)
      return error_text(EX1_message,
                        ism_text);
    expression = Frame[1];
    if (is_SYM(expression))
      return Grammar_True;
    return Grammar_False; }

static VOI_type initialize_is_symbol_native_M(VOI_type)
  { native_define_M(ism_text,
                    is_symbol_native); }

/*--------------------------------- vector? ----------------------------------*/

static const TXT_type ivc_text = "vector?";

static EXP_type is_vector_native(VEC_type Frame)
  { EXP_type expression;
    UNS_type size;
    size = size_VEC(Frame);
    if (size != 1)
      return error_text(EX1_message,
                        ivc_text);
    expression = Frame[1];
    if (is_VEC(expression))
      return Grammar_True;
    return Grammar_False; }

static VOI_type initialize_is_vector_native_M(VOI_type)
  { native_define_M(ivc_text,
                    is_vector_native); }

/*----------------------------------------------------------------------------*/
/*---------------------------------- pairs -----------------------------------*/
/*----------------------------------------------------------------------------*/

/*----------------------------------- cons -----------------------------------*/

static const TXT_type cns_text = "cons";

static EXP_type cons_native_C(VEC_type Frame)
  { EXP_type car,
             cdr;
    UNS_type size;
    size = size_VEC(Frame);
    if (size != 2)
      return error_text(EX2_message,
                        cns_text);
    SURVIVE_DEFAULT_C(Frame);
    car = Frame[1];
    cdr = Frame[2];
    return make_PAI_M(car,
                      cdr); }

static VOI_type initialize_cons_native_M(VOI_type)
  { native_define_M(cns_text,
                    cons_native_C); }

/*----------------------------------- car ------------------------------------*/

static const TXT_type car_text = "car";

static EXP_type car_native(VEC_type Frame)
  { EXP_type value;
    PAI_type pair;
    UNS_type size;
    size = size_VEC(Frame);
    if (size != 1)
      return error_text(EX1_message,
                        car_text);
    pair = Frame[1];
    if (!is_PAI(pair))
      return error_text(NAP_message,
                        car_text);
    value = pair->car;
    return value; }

static VOI_type initialize_car_native_M(VOI_type)
  { native_define_M(car_text,
                    car_native); }

/*----------------------------------- cdr ------------------------------------*/

static const TXT_type cdr_text = "cdr";

static EXP_type cdr_native(VEC_type Frame)
  { EXP_type value;
    PAI_type pair;
    UNS_type size;
    size = size_VEC(Frame);
    if (size != 1)
      return error_text(EX1_message,
                        cdr_text);
    pair = Frame[1];
    if (!is_PAI(pair))
      return error_text(NAP_message,
                        cdr_text);
    value = pair->cdr;
    return value; }

static VOI_type initialize_cdr_native_M(VOI_type)
  { native_define_M(cdr_text,
                    cdr_native); }

/*---------------------------------- length ----------------------------------*/

static const TXT_type lng_text = "length";

static EXP_type length_native(VEC_type Frame)
  { EXP_type value;
    NBR_type number;
    UNS_type length,
             size;
    size = size_VEC(Frame);
    if (size != 1)
      return error_text(EX1_message,
                        lng_text);
    value = Frame[1];
    length = Grammar_List_Length(value,
                                 lng_text);
    number = make_NBR(length);
    return number; }

/*----------------------------------------------------------------------------*/

static VOI_type initialize_length_native_M(VOI_type)
  { native_define_M(lng_text,
                    length_native); }

/*------------------------------------ list ----------------------------------*/

static const TXT_type lst_text = "list";

static EXP_type list_native_C(VEC_type Frame)
  { EXP_type value;
    PAI_type pair;
    UNS_type index,
             size;
    size = size_VEC(Frame);
    if (size == 0)
      return Grammar_Null;
    SURVIVE_DEFAULT_C(Frame);
    value = Frame[size];
    pair = make_PAI_M(value,
                      Grammar_Null);
    Frame[size] = pair;
    for (index = size - 1;
         index > 0;
         index--)
      { SURVIVE_DEFAULT_C(Frame);
        value = Frame[index];
        pair = Frame[index + 1];
        pair = make_PAI_M(value,
                          pair);
        Frame[index] = pair; }
    return pair; }

static VOI_type initialize_list_native_M(VOI_type)
  { native_define_M(lst_text,
                    list_native_C); }

/*--------------------------------- set-car! ---------------------------------*/

static const TXT_type sca_text = "set-car!";

static EXP_type set_car_native(VEC_type Frame)
  { EXP_type value;
    PAI_type pair;
    UNS_type size;
    size = size_VEC(Frame);
    if (size != 2)
      return error_text(EX2_message,
                        sca_text);
    pair  = Frame[1];
    value = Frame[2];
    if (!is_PAI(pair))
      return error_text(NAP_message,
                        sca_text);
    pair->car = value;
    return pair; }

static VOI_type initialize_set_car_native_M(VOI_type)
  { native_define_M(sca_text,
                    set_car_native); }

/*--------------------------------- set-cdr! ---------------------------------*/

static const TXT_type scd_text = "set-cdr!";

static EXP_type set_cdr_native(VEC_type Frame)
  { EXP_type value;
    PAI_type pair;
    UNS_type size;
    size = size_VEC(Frame);
    if (size != 2)
      return error_text(EX2_message,
                        scd_text);
    pair  = Frame[1];
    value = Frame[2];
    if (!is_PAI(pair))
      return error_text(NAP_message,
                        scd_text);
    pair->cdr = value;
    return pair; }

static VOI_type initialize_set_cdr_native_M(VOI_type)
  { native_define_M(scd_text,
                    set_cdr_native); }

/*---------------------------------- assoc -----------------------------------*/

static const TXT_type ass_text = "assoc";

static EXP_type assoc_native(VEC_type Frame)
  { EXP_type value;
    PAI_type list,
             pair;
    TAG_type tag;
    UNS_type size;
    size = size_VEC(Frame);
    if (size != 2)
      return error_text(EX2_message,
                        ass_text);
    value = Frame[1];
    list  = Frame[2];
    for (;;)
      { tag = Grammar_Tag(list);
        switch (tag)
          { case NUL_tag:
              return Grammar_False;
            case PAI_tag:
              pair = list->car;
              if (!is_PAI(pair))
                return error_text(NAP_message,
                                  ass_text);
              if (value == pair->car)
                return pair;
              list = list->cdr;
              break;
            default:
              return error_text(NAL_message,
                                ass_text); }}}

static VOI_type initialize_assoc_native_M(VOI_type)
  { native_define_M(ass_text,
                    assoc_native); }

/*----------------------------------------------------------------------------*/
/*---------------------------------- vectors ---------------------------------*/
/*----------------------------------------------------------------------------*/

/*-------------------------------- make-vector -------------------------------*/

static const TXT_type mkv_text = "make-vector";

static EXP_type make_vector_native_C(VEC_type Frame)
  { EXP_type value;
    NBR_type number;
    SGN_type raw_number;
    UNS_type index,
             size;
    VEC_type vector;
    size = size_VEC(Frame);
    if (size == 0)
      return error_text(AL1_message,
                        mkv_text);
    if (size > 2)
      return error_text(AM2_message,
                        mkv_text);
    value = (size == 1)? Grammar_Unspecified: Frame[2];
    number = Frame[1];
    if (!is_NBR(number))
      return error_text(NAN_message,
                        mkv_text);
    raw_number = get_NBR(number);
    if (raw_number < 0)
      return error_text(NNN_message,
                        mkv_text);
    if (raw_number == 0)
      return Grammar_Empty_Vector_P;
    SURVIVE_DYNAMIC_C(value,
                      raw_number);
    vector = make_VEC_M(raw_number);
    for (index = 1;
         index <= (UNS_type)raw_number;
         index++)
      vector[index] = value;
    return vector; }

static VOI_type initialize_make_vector_native_M(VOI_type)
  { native_define_M(mkv_text,
                    make_vector_native_C); }

/*------------------------------- vector-ref ---------------------------------*/

static const TXT_type vrf_text = "vector-ref";

static EXP_type vector_ref_native(VEC_type Frame)
  { EXP_type value;
    NBR_type number;
    SGN_type raw_number;
    UNS_type size;
    VEC_type vector;
    size = size_VEC(Frame);
    if (size != 2)
      return error_text(EX2_message,
                        vrf_text);
    vector = Frame[1];
    if (!is_VEC(vector))
      return error_text(NAV_message,
                        vrf_text);
    size = size_VEC(vector);
    number = Frame[2];
    if (!is_NBR(number))
      return error_text(NAN_message,
                        vrf_text);
    raw_number = get_NBR(number);
    if (raw_number < 0)
      return error_text(NNN_message,
                        vrf_text);
    if ((UNS_type)raw_number >= size)
      return error_text(IOR_message,
                        vrf_text);
    value = vector[raw_number + 1];
    return value; }

static VOI_type initialize_vector_ref_native_M(VOI_type)
  { native_define_M(vrf_text,
                    vector_ref_native); }

/*------------------------------- vector-set! --------------------------------*/

static const TXT_type vst_text = "vector-set!";

static EXP_type vector_set_native(VEC_type Frame)
  { EXP_type value;
    NBR_type number;
    SGN_type raw_number;
    UNS_type size;
    VEC_type vector;
    size = size_VEC(Frame);
    if (size != 3)
      return error_text(EX3_message,
                        vst_text);
    vector = Frame[1];
    if (!is_VEC(vector))
      return error_text(NAV_message,
                        vst_text);
    size = size_VEC(vector);
    number = Frame[2];
    if (!is_NBR(number))
      return error_text(NAN_message,
                        vst_text);
    raw_number = get_NBR(number);
    if (raw_number < 0)
      return error_text(NNN_message,
                        vst_text);
    if ((UNS_type)raw_number >= size)
      return error_text(IOR_message,
                        vst_text);
    value = Frame[3];
    vector[raw_number + 1] = value;
    return vector; }

static VOI_type initialize_vector_set_native_M(VOI_type)
  { native_define_M(vst_text,
                    vector_set_native); }

/*---------------------------------- vector ----------------------------------*/

static const TXT_type vec_text = "vector";

static EXP_type vector_native(VEC_type Frame)
  { return Frame; }

static VOI_type initialize_vector_native_M(VOI_type)
  { native_define_M(vec_text,
                    vector_native); }

/*------------------------------ vector-length -------------------------------*/

static const TXT_type vcl_text = "vector-length";

static EXP_type vector_length_native(VEC_type Frame)
  { EXP_type value;
    UNS_type size;
    VEC_type vector;
    size = size_VEC(Frame);
    if (size != 1)
      return error_text(EX1_message,
                        vcl_text);
    vector = Frame[1];
    if (!is_VEC(vector))
      return error_text(NAV_message,
                        vcl_text);
    size = size_VEC(vector);
    value = make_NBR(size);
    return value; }

static VOI_type initialize_vector_length_native_M(VOI_type)
  { native_define_M(vcl_text,
                    vector_length_native); }

/*----------------------------------------------------------------------------*/
/*--------------------------------- control ----------------------------------*/
/*----------------------------------------------------------------------------*/

/*---------------------------------- apply -----------------------------------*/

static const TXT_type apl_text = "apply";

static EXP_type apply_native_C(VEC_type Frame)
  { UNS_type size;
    size = size_VEC(Frame);
    if (size != 2)
      return error_text(EX2_message,
                        apl_text);
    return Evaluate_Apply_C(Frame); }

static VOI_type initialize_apply_native_M(VOI_type)
  { native_define_M(apl_text,
                    apply_native_C); }

/*------------------------------------ map -----------------------------------*/

static const TXT_type map_text = "map";

static NBR_type Continue_map;

BEGIN_THREAD_FRAME(mAP);
 THREAD_FRAME_SLOT(VEC, frm);
 THREAD_FRAME_SLOT(PAI, arg);
 THREAD_FRAME_SLOT(PAI, lst);
END_THREAD_FRAME(mAP);

static EXP_type continue_map_C(EXP_type Value)
  { mAP_type map_thread;
    EXP_type value;
    PAI_type arguments,
             list,
             pair;
    TAG_type tag;
    VEC_type frame;
    SURVIVE_DEFAULT_C(Value);
    map_thread = (mAP_type)Thread_Peek();
    arguments = map_thread->arg;
    list      = map_thread->lst;
    tag = Grammar_Tag(arguments);
    list = make_PAI_M(Value,
                      list);
    switch (tag)
      { case NUL_tag:
          Thread_Zap();
          list = Grammar_Reverse_List(list,
                                      map_text);
          return list;
        case PAI_tag:
          value     = arguments->car;
          arguments = arguments->cdr;
          pair = make_PAI_M(value,
                            NULL_PAIR);
          frame = map_thread->frm;
          frame[2] = pair;
          map_thread = (mAP_type)Thread_Keep();
          map_thread->arg = arguments;
          map_thread->lst = list;
          return Evaluate_Apply_C(frame);
        default:
          return error_text(NAL_message,
                            map_text); }}

/*----------------------------------------------------------------------------*/

static EXP_type map_native_C(VEC_type Frame)
  { mAP_type map_thread;
    EXP_type procedure,
             value;
    PAI_type arguments,
             pair;
    TAG_type tag;
    UNS_type size;
    VEC_type frame;
    size = size_VEC(Frame);
    if (size != 2)
      return error_text(EX2_message,
                        map_text);
    SURVIVE_DEFAULT_C(Frame);
    arguments = Frame[2];
    tag = Grammar_Tag(arguments);
    switch (tag)
      { case NUL_tag:
          return NULL_PAIR;
        case PAI_tag:
          procedure = Frame[1];
          value     = arguments->car;
          arguments = arguments->cdr;
          frame = Environment_Create_Frame_M(2);
          pair = make_PAI_M(value,
                            NULL_PAIR);
          frame[1] = procedure;
          frame[2] = pair;
          map_thread = (mAP_type)Thread_Push_M(Continue_map,
                                               mAP_size);
          map_thread->frm = frame;
          map_thread->arg = arguments;
          map_thread->lst = NULL_PAIR;
          return Evaluate_Apply_C(frame);
        default:
          return error_text(NAL_message,
                            map_text); }}

static VOI_type initialize_map_native_M(VOI_type)
  { Continue_map = Thread_Register(continue_map_C);
    native_define_M(map_text,
                    map_native_C); }

/*----------------------------------- eval -----------------------------------*/

static const TXT_type evl_text = "eval";

static NBR_type Continue_eval;

BEGIN_THREAD_FRAME(eVA);
 THREAD_FRAME_SLOT(VEC, env);
 THREAD_FRAME_SLOT(VEC, frm);
END_THREAD_FRAME(eVA);

static EXP_type continue_eval(EXP_type Value)
  { eVA_type eval_thread;
    VEC_type environment,
             frame;
    eval_thread = (eVA_type)Thread_Peek();
    environment = eval_thread->env;
    frame       = eval_thread->frm;
    Thread_Zap();
    Environment_Set_Environment_And_Frame(environment,
                                          frame);
    return Value; }

static EXP_type eval_native_C(VEC_type Frame)
  { eVA_type eval_thread;
    EXP_type compiled_expression,
             expression;
    UNS_type size;
    size = size_VEC(Frame);
    if (size != 1)
      return error_text(EX1_message,
                        evl_text);
    expression = Frame[1];
    compiled_expression = Compile_C(expression);
    SURVIVE_DEFAULT_C(compiled_expression);
    eval_thread = (eVA_type)Thread_Push_M(Continue_eval,
                                          eVA_size);
    eval_thread->env = Environment_Get_Environment();
    eval_thread->frm = Environment_Get_Frame();
    Environment_Reset();
    return Evaluate_C(compiled_expression); }

static VOI_type initialize_eval_native_M(VOI_type)
  { Continue_eval = Thread_Register(continue_eval);
    native_define_M(evl_text,
                    eval_native_C); }

/*----------------------------------------------------------------------------*/
/*----------------------------------- input ----------------------------------*/
/*----------------------------------------------------------------------------*/

/*----------------------------------- read -----------------------------------*/

static const TXT_type rdd_text = "read";

static EXP_type read_native_C(VEC_type Frame)
  { EXP_type expression;
    STR_type path;
    TXT_type input,
             text;
    UNS_type size;
    size = size_VEC(Frame);
    if (size > 1)
      return error_text(AM1_message,
                        rdd_text);
    if (size == 0)
      input = Main_Read();
    else
      { path = Frame[1];
        if (!is_STR(path))
          return error_text(NAS_message,
                            rdd_text);
        text = path->txt;
        input = Main_Load(text); }
    expression = Read_C(input);
    return expression; }

static VOI_type initialize_read_native_M(VOI_type)
  { native_define_M(rdd_text,
                    read_native_C); }

/*----------------------------------------------------------------------------*/
/*---------------------------------- output ----------------------------------*/
/*----------------------------------------------------------------------------*/

/*--------------------------------- display ----------------------------------*/

static const TXT_type dis_text = "display";

static EXP_type display_native(VEC_type Frame)
  { EXP_type value;
    UNS_type size;
    size = size_VEC(Frame);
    if (size != 1)
      return error_text(EX1_message,
                        dis_text);
    value = Frame[1];
    Print_Display(value);
    return Grammar_Unspecified; }

static VOI_type initialize_display_native_M(VOI_type)
  { native_define_M(dis_text,
                    display_native); }

/*--------------------------------- newline ----------------------------------*/

static const TXT_type nwl_text = "newline";

static EXP_type newline_native(VEC_type Frame)
  { UNS_type size;
    size = size_VEC(Frame);
    if (size != 0)
      return error_text(EX0_message,
                        nwl_text);
    Print_Newline();
    return Grammar_Unspecified; }

static VOI_type initialize_newline_native_M(VOI_type)
  { native_define_M(nwl_text,
                    newline_native); }

/*---------------------------------- pretty ----------------------------------*/

static const TXT_type prt_text = "pretty";

static EXP_type pretty_native(VEC_type Frame)
  { EXP_type compiled_expression,
             expression;
    UNS_type size;
    size = size_VEC(Frame);
    if (size != 1)
      return error_text(EX1_message,
                        prt_text);
    expression = Frame[1];
    compiled_expression = Compile_C(expression);
    Print_Newline();
    Print_Print(compiled_expression);
    Print_Newline();
    return Grammar_Unspecified; }

static VOI_type initialize_pretty_native_M(VOI_type)
  { native_define_M(prt_text,
                    pretty_native); }

/*----------------------------------------------------------------------------*/
/*---------------------------------- service ---------------------------------*/
/*----------------------------------------------------------------------------*/

/*----------------------------------- clock ----------------------------------*/

static const TXT_type clk_text = "clock";

static EXP_type clock_native_C(VEC_type Frame)
  { FLO_type raw_real;
    REA_type real;
    UNS_type size,
             ticks;
    size = size_VEC(Frame);
    if (size != 0)
      return error_text(EX0_message,
                        clk_text);
    ticks = clock();
    raw_real = (FLO_type)ticks / CLOCKS_PER_SEC;
    Grammar_Default_Claim_C();
    real = make_REA_M(raw_real);
    return real; }

static VOI_type initialize_clock_native_M(VOI_type)
  { native_define_M(clk_text,
                    clock_native_C); }

/*---------------------------------- collect ---------------------------------*/

static const TXT_type col_text = "collect";

static EXP_type collect_native_C(VEC_type Frame)
  { UNS_type size;
    size = size_VEC(Frame);
    if (size != 0)
      return error_text(EX0_message,
                        col_text);
    Grammar_Reclaim_C();
    return Grammar_Unspecified; }

static VOI_type initialize_collect_native_M(VOI_type)
  { native_define_M(col_text,
                    collect_native_C); }

/*----------------------------------- exit -----------------------------------*/

static const TXT_type exi_text = "exit";

static EXP_type exit_native(VEC_type Frame)
  { UNS_type size;
    size = size_VEC(Frame);
    if (size != 0)
      return error_text(EX0_message,
                        exi_text);
    Main_Terminate();
    return Grammar_Unspecified; }

static VOI_type initialize_exit_native_M(VOI_type)
  { native_define_M(exi_text,
                    exit_native); }

/*--------------------------------- random -----------------------------------*/

static const TXT_type random_text = "random";

static NBR_type random_native(VEC_type Frame)
  { NBR_type number;
    UNS_type random,
             size;
        size = size_VEC(Frame);
    if (size != 0)
      return error_text(EX0_message,
                        random_text);
    random = rand();
    random >>= 2;
    number = make_NBR(random);
    return number; }

/*----------------------------------------------------------------------------*/

static VOI_type initialize_random_native_M(VOI_type)
  { native_define_M(random_text,
                    random_native); }

/*----------------------------------- run ------------------------------------*/

static NBR_type Continue_print;

static EXP_type continue_print(EXP_type Value)
  { Print_Print(Value);
    Main_Proceed();
    return Grammar_Unspecified; }

/*----------------------------------------------------------------------------*/

static VOI_type read_eval_print_C(TXT_type Input)
  { CCC_type function_C;
    EXP_type compiled_expression,
             expression,
             value;
    Dictionary_Checkpoint();
    expression = Read_C(Input);
    compiled_expression = Compile_C(expression);
    SURVIVE_DEFAULT_C(compiled_expression);
    Thread_Push_M(Continue_print,
                  0);
    value = Evaluate_C(compiled_expression);
    for (;;)
      { function_C = Thread_Retrieve();
        value = function_C(value); }}

static VOI_type initialize_run(VOI_type)
  { Continue_print = Thread_Register(continue_print); }

/*----------------------------- public functions -----------------------------*/

VOI_type Native_Reset_C(VOI_type)
  { UNS_type global_frame_usage;
    Stack_Reset();
    Dictionary_Rollback();
    global_frame_usage = Dictionary_Get_Frame_Size();
    Environment_Rollback(global_frame_usage);
    Thread_Clear(); }

VOI_type Native_Run_C(TXT_type Input)
  { read_eval_print_C(Input); }

/*------------------------------ initialization ------------------------------*/

VOI_type Native_Initialize_M(ADR_type Memory,
                             UNS_type Size)
  { Grammar_Initialize_M(Memory,
                         Size);

    Dictionary_Initialize();
    Environment_Initialize_M();
    Pool_Initialize_M();
    Stack_Initialize_M();
    Thread_Initialize();

    Compile_Initialize();
    Evaluate_Initialize();
    Print_Initialize();
    Read_Initialize();

    initialize_add_native_M();
    initialize_apply_native_M();
    initialize_assoc_native_M();
    initialize_car_native_M();
    initialize_cdr_native_M();
    initialize_clock_native_M();
    initialize_collect_native_M();
    initialize_cons_native_M();
    initialize_display_native_M();
    initialize_divide_native_M();
    initialize_equal_native_M();
    initialize_equivalent_native_M();
    initialize_eval_native_M();
    initialize_exit_native_M();
    initialize_greater_native_M();
    initialize_greater_or_equal_native_M();
    initialize_is_null_native_M();
    initialize_is_pair_native_M();
    initialize_is_procedure_native_M();
    initialize_is_real_native_M();
    initialize_is_symbol_native_M();
    initialize_is_vector_native_M();
    initialize_less_native_M();
    initialize_less_or_equal_native_M();
    initialize_length_native_M();
    initialize_list_native_M();
    initialize_make_vector_native_M();
    initialize_map_native_M();
    initialize_multiply_native_M();
    initialize_newline_native_M();
    initialize_not_native_M();
    initialize_pretty_native_M();
    initialize_quotient_native_M();
    initialize_random_native_M();
    initialize_read_native_M();
    initialize_remainder_native_M();
    initialize_set_car_native_M();
    initialize_set_cdr_native_M();
    initialize_sqrt_native_M();
    initialize_subtract_native_M();
    initialize_vector_length_native_M();
    initialize_vector_native_M();
    initialize_vector_ref_native_M();
    initialize_vector_set_native_M();

    initialize_run(); }
